@@ -16,10 +16,10 @@ Versioning semantics:
 
 __intname__ = "command_runner"
 __author__ = "Orsiris de Jong"
-__copyright__ = "Copyright (C) 2015-2021 Orsiris de Jong"
+__copyright__ = "Copyright (C) 2015-2022 Orsiris de Jong"
 __licence__ = "BSD 3 Clause"
-__version__ = "1.2.0"
-__build__ = "2021090701"
+__version__ = "1.3.1"
+__build__ = "2022041601"
 
 import io
 import os
@@ -32,10 +32,13 @@ from time import sleep
 
 try:
     import psutil
-    import signal
 except ImportError:
     # Don't bother with an error since we need command_runner to work without dependencies
     pass
+try:
+    import signal
+except ImportError:
+    pass
 
 # Python 2.7 compat fixes (queue was Queue)
 try:
@@ -50,6 +53,7 @@ try:
 except ImportError:
     pass
 try:
+    # pylint: disable=E0601 (used-before-assignment)
     FileNotFoundError
 except NameError:
     # pylint: disable=W0622 (redefined-builtin)
@@ -140,18 +144,22 @@ def kill_childs_mod(
         A ValueError will be raised in any other case. Note that not all systems define the same set of signal names;
         an AttributeError will be raised if a signal name is not defined as SIG* module level constant.
         """
-        if not soft_kill and hasattr(signal, "SIGKILL"):
-            # Don't bother to make pylint go crazy on Windows
-            # pylint: disable=E1101
-            sig = signal.SIGKILL
-        else:
-            sig = signal.SIGTERM
+        try:
+            if not soft_kill and hasattr(signal, "SIGKILL"):
+                # Don't bother to make pylint go crazy on Windows
+                # pylint: disable=E1101
+                sig = signal.SIGKILL
+            else:
+                sig = signal.SIGTERM
+        except NameError:
+            sig = None
     ### END COMMAND_RUNNER MOD
 
-    def _process_killer(process,  # type: Union[subprocess.Popen, psutil.Process]
-                        sig,  # type: signal.valid_signals
-                        soft_kill  # type: bool
-        ):
+    def _process_killer(
+        process,  # type: Union[subprocess.Popen, psutil.Process]
+        sig,  # type: signal.valid_signals
+        soft_kill,  # type: bool
+    ):
         # (...) -> None
         """
         Simple abstract process killer that works with signals in order to avoid reused PID race conditions
@@ -160,7 +168,9 @@ def kill_childs_mod(
         if sig:
             try:
                 process.send_signal(sig)
-            except psutil.NoSuchProcess:
+            # psutil.NoSuchProcess might not be available, let's be broad
+            # pylint: disable=W0703
+            except Exception:
                 pass
         else:
             if soft_kill:
@@ -169,18 +179,21 @@ def kill_childs_mod(
                 process.kill()
 
     try:
-        parent = psutil.Process(pid if pid is not None else os.getpid())
-    except psutil.NoSuchProcess:
-        return False
-    except NameError:
-        os.kill(pid, sig)
+        current_process = psutil.Process(pid if pid is not None else os.getpid())
+    # psutil.NoSuchProcess might not be available, let's be broad
+    # pylint: disable=W0703
+    except Exception:
+        if itself:
+            os.kill(
+                pid, 15
+            )  # 15 being signal.SIGTERM or SIGKILL depending on the platform
         return False
 
-    for child in parent.children(recursive=True):
+    for child in current_process.children(recursive=True):
         _process_killer(child, sig, soft_kill)
 
     if itself:
-        _process_killer(parent, sig, soft_kill)
+        _process_killer(current_process, sig, soft_kill)
     return True
 
 
@@ -190,8 +203,8 @@ def command_runner(
     timeout=3600,  # type: Optional[int]
     shell=False,  # type: bool
     encoding=None,  # type: Optional[str]
-    stdout=None,  # type: Union[int, str]
-    stderr=None,  # type: Union[int, str]
+    stdout=None,  # type: Optional[Union[int, str]]
+    stderr=None,  # type: Optional[Union[int, str]]
     windows_no_window=False,  # type: bool
     live_output=False,  # type: bool
     method="monitor",  # type: str
@@ -258,25 +271,28 @@ def command_runner(
 
     # Decide whether we write to output variable only (stdout=None), to output variable and stdout (stdout=PIPE)
     # or to output variable and to file (stdout='path/to/file')
+    stdout_to_file = False
     if stdout is None:
         _stdout = PIPE
-        stdout_to_file = False
     elif isinstance(stdout, str):
         # We will send anything to file
         _stdout = open(stdout, "wb")
         stdout_to_file = True
+    elif stdout is False:
+        _stdout = subprocess.DEVNULL
     else:
         # We will send anything to given stdout pipe
         _stdout = stdout
-        stdout_to_file = False
 
     # The only situation where we don't add stderr to stdout is if a specific target file was given
+    stderr_to_file = False
     if isinstance(stderr, str):
         _stderr = open(stderr, "wb")
         stderr_to_file = True
+    elif stderr is False:
+        _stderr = subprocess.DEVNULL
     else:
         _stderr = subprocess.STDOUT
-        stderr_to_file = False
 
     def to_encoding(
         process_output,  # type: Union[str, bytes]
@@ -385,6 +401,9 @@ def command_runner(
             # output_queue has finished sending data, so we catch the exit code
             while process.poll() is None:
                 __check_timeout(begin_time, timeout)
+            # Additional timeout check to make sure we don't return an exit code from processes
+            # that were killed because of timeout
+            __check_timeout(begin_time, timeout)
             exit_code = process.poll()
             return exit_code, output
 
@@ -394,7 +413,7 @@ def command_runner(
     def _timeout_check_thread(
         process,  # type: Union[subprocess.Popen[str], subprocess.Popen]
         timeout,  # type: int
-        timeout_dict,  # type: dict
+        timeout_queue,  # type: queue.Queue
     ):
         # type: (...) -> None
 
@@ -406,7 +425,7 @@ def command_runner(
         while True:
             if timeout and (datetime.now() - begin_time).total_seconds() > timeout:
                 kill_childs_mod(process.pid, itself=True, soft_kill=False)
-                timeout_dict["is_timeout"] = True
+                timeout_queue.put(True)
                 break
             if process.poll() is not None:
                 break
@@ -424,12 +443,16 @@ def command_runner(
         Get stdout output and return it
         """
 
-        # Let's create a mutable object since it will be shared with a thread
-        timeout_dict = {"is_timeout": False}
+        # Shared mutable objects have proven to have race conditions with PyPy 3.7 (mutable object
+        # is changed in thread, but outer monitor function has still old mutable object state)
+        # Strangely, this happened only sometimes on github actions/ubuntu 20.04.3 & pypy 3.7
+        # Let's create a queue to get the timeout thread response on a deterministic way
+        timeout_queue = queue.Queue()
+        is_timeout = False
 
         thread = threading.Thread(
             target=_timeout_check_thread,
-            args=(process, timeout, timeout_dict),
+            args=(process, timeout, timeout_queue),
         )
         thread.setDaemon(True)
         thread.start()
@@ -442,27 +465,41 @@ def command_runner(
             # Also it won't allow communicate() to get incomplete output on timeouts
             while process.poll() is None:
                 sleep(MIN_RESOLUTION)
-                if timeout_dict["is_timeout"]:
+                try:
+                    is_timeout = timeout_queue.get_nowait()
+                except queue.Empty:
+                    pass
+                else:
                     break
-
                 # We still need to use process.communicate() in this loop so we don't get stuck
                 # with poll() is not None even after process is finished
+                if _stdout is not False:
+                    try:
+                        stdout, _ = process.communicate()
+                    # ValueError is raised on closed IO file
+                    except (TimeoutExpired, ValueError):
+                        pass
+            exit_code = process.poll()
+
+            if _stdout is not False:
                 try:
                     stdout, _ = process.communicate()
-                # ValueError is raised on closed IO file
                 except (TimeoutExpired, ValueError):
                     pass
-
-            exit_code = process.poll()
-            try:
-                stdout, _ = process.communicate()
-            except (TimeoutExpired, ValueError):
-                pass
             process_output = to_encoding(stdout, encoding, errors)
 
-            if timeout_dict["is_timeout"]:
-                raise TimeoutExpired(process, timeout, process_output)
+            # On PyPy 3.7 only, we can have a race condition where we try to read the queue before
+            # the thread could write to it, failing to register a timeout.
+            # This workaround prevents reading the queue while the thread is still alive
+            while thread.is_alive():
+                sleep(MIN_RESOLUTION)
 
+            try:
+                is_timeout = timeout_queue.get_nowait()
+            except queue.Empty:
+                pass
+            if is_timeout:
+                raise TimeoutExpired(process, timeout, process_output)
             return exit_code, process_output
         except KeyboardInterrupt:
             raise KbdInterruptGetOutput(process_output)
@@ -506,7 +543,7 @@ def command_runner(
             )
 
         try:
-            if method == "poller" or live_output:
+            if method == "poller" or live_output and _stdout is not False:
                 exit_code, output = _poll_process(process, timeout, encoding, errors)
             else:
                 exit_code, output = _monitor_process(process, timeout, encoding, errors)
